#ifdef VIODS_SUPPORT

// This is the implementation of the server side control functions
//
// Test programm included at the end -DTEST_MAIN=1 to enable it
//
// NB: no locking is done
//
// Author: NFSP team
// Date: 2002/07
// Licence: GPLv2

#include "nfsp_viods.h"

#include "logging.h"
#include "default.h"	/* for MAX_IODS and MAX_VIODS */
#include "list.h"

#include <string.h>

#ifdef TEST_MAIN
#undef MAX_IODS
#define MAX_IODS 9
#undef MAX_VIODS
#define MAX_VIODS 3
#endif


typedef struct phys_iod_t {
	struct list_head list;
	struct sockaddr_in addr;
	char used;			// boolean 0 or 1
	char bound;			// boolean 0 or 1
	int iviod;			// only valid if bound, index of related viod
} phys_iod_t;

#define LOCK do {} while(0)
#define UNLOCK  do {} while(0)

static phys_iod_t iods[MAX_IODS];
static phys_iod_t *viods[MAX_VIODS];
static int nfspd_nbviods;

nfspd_state_t nfspd_state;

/* list of unused iods */
static struct list_head unused;

// -----------------------------------------------

static int iodp2ih(const phys_iod_t *iodp) {
	if (iodp < iods || iodp > iods + MAX_IODS) {
		return -1;
	}
	return (iodp - iods);
}

static phys_iod_t* ih2iodp(const int ih) {
	if ((ih < 0) || (ih >= MAX_IODS)) {
		return NULL;
	}
	return iods + ih;
}

static phys_iod_t* get_unused() {
	phys_iod_t *iod;

	if (unused.next == &unused) {
		return NULL; /* no more free descriptors */
	}
	iod = list_entry(unused.next, phys_iod_t, list);
	list_del(&iod->list); 

	if (iod->used) {
		Dprintf(L_WARNING, "get_unused(): oops - iod already used");
		return NULL;
	}
	if (iod->bound) {
		Dprintf(L_WARNING, "get_unused(): oops - iod already bound");
		return NULL;
	}
	iod->used = 1;
	return iod;
}

static int put_unused(phys_iod_t *iod) {
	if (!iod) {
		Dprintf(L_WARNING, "put_unused(): oops - iod is null");
		return -1;
	}
	if (!iod->used) {
		Dprintf(L_WARNING, "put_unused(): oops - iod is not used");
		return -1;
	}
	if (iod->bound) {
		Dprintf(L_WARNING, "put_unused(): oops - iod is still bound");
		return -1;
	}
	list_add(&iod->list, &unused);
	
	iod->used = 0;
	return 0;
}


// -----------------------------------------------


int nfspd_init_svc() {
	int i;

	memset(iods, 0, MAX_IODS * sizeof(phys_iod_t));
	memset(viods, 0, MAX_VIODS * sizeof(phys_iod_t*));
	nfspd_nbviods = 0;
	nfspd_state = NFSPD_DROPPING_MODE;

	INIT_LIST_HEAD(&unused);
	for(i=0; i < MAX_IODS; i++) {
		list_add_tail(&iods[i].list, &unused);
	}

	return 0;
}


int nfspd_addiod_svc(struct sockaddr_in *sinp) {
	phys_iod_t *iod;
	int err;
	int i, ih = -1;

	LOCK;

	if (sinp == NULL) {
		ih = -EINVAL;
		goto out_release;
	}
	
	for (i = 0; i < MAX_IODS; i++) {
		if (iods[i].used && 
			sinp->sin_addr.s_addr == iods[i].addr.sin_addr.s_addr &&
			sinp->sin_port == iods[i].addr.sin_port) {
			ih = -EEXIST;
			goto out_release;
		}
	}

	iod = get_unused();
	if (!iod) {
		err = -ENOMEM;
		goto out_release;
	}
	ih = iodp2ih(iod);

	memcpy(&iod->addr, sinp, sizeof(iod->addr));

out_release:
	UNLOCK;
	return ih;
}

int nfspd_freeiod_svc(int ih) {
	phys_iod_t *iod;
	int err;
	
	LOCK;
	iod = ih2iodp(ih);
	err = -EINVAL;
	if ((!iod) || (!iod->used) || (iod->bound)) {
		goto out_release;
	}
	list_del(&(iod->list)); 
	err = put_unused(iod);
		
out_release:
	UNLOCK;
	return err;
}

/* check from here*/
int nfspd_bindiod_svc(int iviod, int ih) {
	phys_iod_t *iodp;
	int err;

	err = -EINVAL;
	if ((iviod < 0) || (iviod >= MAX_VIODS)) goto out;

	LOCK;
	iodp = ih2iodp(ih);
	if ((!iodp) || (!iodp->used) || (iodp->bound)) {
		goto out_release;
	}

	err = 0;
	if (!viods[iviod]) {
		/* no iod already bound to this viod */
		INIT_LIST_HEAD(&iodp->list);
		viods[iviod] = iodp;
	} else {
		/* simply add to the list */
		list_add_tail(&iodp->list, &(viods[iviod]->list));
	}

	iodp->bound = 1;
	iodp->iviod = iviod;

out_release:
	UNLOCK;

out:
	return err;
}


int nfspd_unbindiod_svc(int ih) {
	phys_iod_t *iodp;
	int err;

	LOCK;
	iodp = ih2iodp(ih);
	err = -EINVAL;
	if ((!iodp) || (!iodp->used) || (!iodp->bound)) {
		goto out_release;
	}

	if (iodp->bound) {
		int iviod = iodp->iviod;

		if (viods[iviod]==iodp){ // this iod correspond to the viod ptr
			if (iodp->list.next == &iodp->list) {	/* this iod is the last */
				viods[iviod] = NULL;
			} 
			else { // the viods[iviod] have to be assign on the next iod
				// was: viods[iviod]=iodp->list.next;
				viods[iviod] = list_entry(viods[iviod]->list.next, struct phys_iod_t, list);
			}
		
		}
	}
	list_del(&iodp->list);
      
	err = 0;
	iodp->bound = 0;

out_release:
	UNLOCK;
	return err;
}


int nfspd_setnbviods_svc(int n) {
	int err;

	LOCK;
	Dprintf(D_NFSP_VIODS, "nfspd_setnbviods_svc = %d", n);
	if (n < 1 || n > MAX_VIODS) {
		err = -EINVAL;
		goto out_release;
	}
	if (nfspd_state != NFSPD_DROPPING_MODE) {
		err = -EBUSY;
		goto out_release;
	}
	nfspd_nbviods = n;
	err = 0;
	
out_release:
	UNLOCK;

	return err;
}

int nfspd_getnbviods_svc() {
	int err;

	LOCK;
	err = nfspd_nbviods;
	UNLOCK;

	return err;
}


int nfspd_drop_svc() {
	LOCK;
	nfspd_state = NFSPD_DROPPING_MODE;
	UNLOCK;
	return 0;
}


int nfspd_forward_svc() {
	LOCK;
	if (nfspd_nbviods <= 0) {
		return -EPERM;
	}
	nfspd_state = NFSPD_FORWARDING_MODE;
	UNLOCK;
	return 0;
}

int nfspd_flush_svc() {
	LOCK;
	nfspd_init_svc();
	UNLOCK;
	return 0;
}

int nfspd_get_viod_size_svc(int iviod) {
	int err;

	LOCK;
	if (iviod < 0 || iviod >= nfspd_nbviods) {
		err = -EINVAL;
		goto out_release;
	}

	err = (viods[iviod] != NULL);

out_release:
	UNLOCK;
	return err;
}


struct sockaddr_in * get_iod(int iviod) {
	struct sockaddr_in * ret;
	struct phys_iod_t *prev;

	prev = viods[iviod];

	ret = &viods[iviod]->addr;
	
	viods[iviod] = list_entry(viods[iviod]->list.next, struct phys_iod_t, list);
	
	Dprintf(D_NFSP_VIODS, "get_iod: iviod=%d was iod %p/ih=%d (returned value) is now iod %p/ih=%d",
			iviod, prev, iodp2ih(prev), viods[iviod], iodp2ih(viods[iviod]));

	return ret;
}

#ifdef WRITEMIRROR_SUPPORT
// Used to fill the nextiods[] field of an iod write mirror request
//
// If iod_tab != NULL, fills iod_tab with the iods of the viod #iviod
// Always return the number of iods in the viod group #iviod
//
// and copy all list content inside sockaddr_in iodTab[MAX_IODS]
//
// WARNING:
//  DO NOT expect anything about the order of the list returned in iod_tab
//  (it may change between two calls !)
int get_all_iod(int iviod, struct sockaddr_in* iod_tab) {
	struct list_head *ptr_index, *ptr_temp; 	// Use to parse the linked list 
	struct phys_iod_t *phys_iod_ptr_temp;
	int iods_number = 0; 			// iod's number inside the viod specified by iviod

	if (iod_tab) {
		Dprintf(D_NFSP_VIODS, "get_all_iod(): The viod %d gathers:\n", iviod);

		// Walks the linked list and copy the sockaddr_in
		list_for_each_safe(ptr_index, ptr_temp, &(viods[iviod]->list)) {
			if (iods_number > MAX_NEXTIODS - 1) {
				Dprintf(L_WARNING, "Too many iods in this viod to fill iod_tab "
						"(must be <= MAX_NEXTIODS=%d)", MAX_NEXTIODS);
				return -1;
			}
			// Retrieve the next iod space and copy
			phys_iod_ptr_temp=list_entry(ptr_index, struct phys_iod_t, list);
			memcpy(&(iod_tab[iods_number]), &(phys_iod_ptr_temp->addr), sizeof(phys_iod_ptr_temp->addr));
			
			Dprintf(D_NFSP_VIODS, "get_all_iod: the iod " NTODSTR "\n",NTOD(iod_tab[iods_number]));
			iods_number++;
		}
	} else {
		// Walks the linked list
		list_for_each_safe(ptr_index, ptr_temp, &(viods[iviod]->list)) {	
			iods_number++;
		
		}
	}
	Dprintf(D_NFSP_VIODS, "get_all_iod: There are %d iod(s) in the viod #%d", iods_number, iviod);
	return iods_number;
}
#endif /* WRITEMIRROR_SUPPORT */



#ifdef SHOWINFO
#include <stdio.h>
#include <stdlib.h>
#include <arpa/inet.h>
void nfspd_viods_showall() {
	int i;
	struct phys_iod_t *iodp;
#ifdef TEST_MAIN
	struct list_head *lh;
#endif /* TEST_MAIN */

	printf("-----------------------------------------------------\n");
	printf("state: %d (%s)\n", nfspd_state, 
		(nfspd_state==NFSPD_DROPPING_MODE) ? "drop" : "fwd" );
	printf("nbviods: %d\n", nfspd_nbviods);

#ifdef TEST_MAIN
	
	printf("\n\nunused list\n");
	i=0;
	lh = &unused;
	printf("  unused[%d]:@=%p/ih=%d\n", i++, iodp, iodp2ih(iodp));
	for (lh = lh->next;
			lh != &unused;
			lh = lh->next 
			) {
		iodp = list_entry(lh, struct phys_iod_t, list);
		printf("  unused[%d]:@=%p/ih=%d\n", i++, iodp, iodp2ih(iodp));
	}

	printf("\n\niods array\n");
	for (i = 0; i < MAX_IODS; i++) {
		printf("iod[%d] ih=%d used=%d bound=%d iviod=%d\n", i,
				iodp2ih(&iods[i]), iods[i].used, iods[i].bound,
				iods[i].iviod);
	}
#endif /* TEST_MAIN */

	printf("\n\nviods array\n");
	for(i = 0; i < MAX_VIODS; i++) {
		// Used to walk the linked list 
		struct list_head *ptr_index, *ptr_temp;

		printf("viods[%d]=%p\n", i, viods[i]);
		if (!viods[i]) {
			continue;
		}

		// ~= (iodp = list_entry(viods[i]->list, struct phys_iod_t, list)
		iodp = viods[i] ;

		printf("  viods[%d]: @=%p ih=%d ip=0x%8.8x port=%d\n", i, iodp,
				iodp2ih(iodp),
				ntohl(iodp->addr.sin_addr.s_addr),
				ntohs(iodp->addr.sin_port));

		list_for_each_safe(ptr_index, ptr_temp, &(viods[i]->list)) {
			iodp = list_entry(ptr_index, struct phys_iod_t, list);
			printf("  viods[%d]: @=%p ih=%d ip=0x%8.8x port=%d\n",
		       i, iodp, iodp2ih(iodp), ntohl(iodp->addr.sin_addr.s_addr),
		       ntohs(iodp->addr.sin_port));
		}
	}

	

	printf("-----------------------------------------------------\n");
}
#endif /* SHOWINFO */

#ifdef TEST_MAIN

int main(int argc, char *argv[]) {
	int ret;

	printf("initialization\n");
	ret = nfspd_init_svc();
	printf("status = %d\n", ret);


	while(1) {
		char line[1024];
		char buf0[1000], buf1[1000], buf2[1000];

		line[0] = buf0[0] = buf1[0] = buf2[0] = '\0';

		ret = -1;
		
		gets(line);
		ret = sscanf(line, "%s %s %s", buf0, buf1, buf2);
		printf("\nGot %d args [%s] [%s] [%s]\n", ret, buf0, buf1, buf2);
		
		switch (buf0[0]) {
			case 'a':
				do {
					struct sockaddr_in sin;

					sin.sin_addr.s_addr = inet_addr(buf1);
					sin.sin_port = htons(atoi(buf2));
					ret = nfspd_addiod_svc(&sin);
				
				  } while (0);
					  break;
			case 'f':
				ret = nfspd_freeiod_svc(atoi(buf1));
				break;
			case 'b':
				ret = nfspd_bindiod_svc(atoi(buf1),atoi(buf2));
				break;
			case 'u':
				ret = nfspd_unbindiod_svc(atoi(buf1));
				break;
			case 's':
				ret = nfspd_setnbviods_svc(atoi(buf1));
				break;
			case 'g':
				ret = nfspd_getnbviods_svc();
				break;
			case 'v':
				ret = nfspd_get_viod_size(atoi(buf1));
				break;
			case 'q':
				exit(0);
				break;
			default:
				printf("option was [%c]", buf0[0]);
		}
		printf("status = %d\n", ret);
		nfspd_viods_showall();
	}

	return 0;
}
#endif /* TEST_MAIN */

#endif /* VIODS_SUPPORT */

