// udpspoof
// defines the spoofing functions for UDP and ethernet packets
//
// Changelog:
//   200205xx Adrien Lebre <adrien.lebre@imag.fr>
//   ETHERNET_send() added
//
// Author:
//   2001xxxx Pierre Lombard <pierre.lombard@imag.fr>
//
// Licence: GPL

#include "udpspoofng.h"
#include "in_cksum.h"		// IP checksumming functions

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <net/if.h>
#include <netinet/udp.h>
#include <arpa/inet.h>
#include <pthread.h>

// use for the PF_PACKET socket and the struct sockaddr_ll (see man 7 packet...).
#include <features.h>    /* for the glibc version number */
#if __GLIBC__ >= 2 && __GLIBC_MINOR >= 1
#  include <netpacket/packet.h>
#  include <net/ethernet.h>     /* the L2 protocols */
#else
#  include <asm/types.h>
#  include <linux/if_packet.h>
#  include <linux/if_ether.h>   /* The L2 protocols */
#endif


#if !defined(_REENTRANT)
# error "_REENTRANT must be defined for multithreading support"
#endif

#include "logging.h"

// Some flavors of gcc/distributions/... don't define those types so they're
// forced here (check /usr/include/netinet/in.h for availability)
#define in_addr_t uint32_t
#define in_port_t uint16_t

typedef struct sockaddr_in sockaddr_in_t;
typedef struct sockaddr sockaddr_t;

unsigned int eth0addr;

#undef IPID_MUTEX

static unsigned short int ipid = 0x6789;
#if IPID_MUTEX
static pthread_mutex_t ipid_mutex = PTHREAD_MUTEX_INITIALIZER;
#endif /* IPID_MUTEX */

static inline unsigned short get_ipid() {
	unsigned short int tmp;
#if IPID_MUTEX
	int ret;
	ret = pthread_mutex_lock(&ipid_mutex);
	if (ret) {
		Dprintf(L_WARNING, "pthread_mutex_lock failed");
	}
	tmp = ipid;
	++ipid;
	ret = pthread_mutex_unlock(&ipid_mutex);
	if (ret) {
		Dprintf(L_WARNING, "pthread_mutex_unlock failed");
	}
#else /* IPID_MUTEX */
	tmp = ipid;
	ipid++;
#endif /* !IPID_MUTEX */
	return tmp;
}

void set_ipid(unsigned short const x) {
#if IPID_MUTEX
	int ret;
	ret = pthread_mutex_lock(&ipid_mutex);
	if (ret) {
		Dprintf(L_WARNING, "pthread_mutex_lock failed");
	}
	ipid = x;
	ret = pthread_mutex_unlock(&ipid_mutex);
	if (ret) {
		Dprintf(L_WARNING, "pthread_mutex_unlock failed");
	}
#else /* IPID_MUTEX */
	ipid = x;
#endif /* !IPID_MUTEX */
}

typedef struct {
	char ifname[IFNAMSIZ];	// name (eth0, lo, etc...)
	in_addr_t ifaddr;	// IP address
	unsigned char ifhwaddr[8];
	ushort ifmtu;		// MTU
	int ifindex;		// interface index (required for the ETHERNET_send())
} my_if_t;

static my_if_t if_lo;
static my_if_t if_eth0;

// itf  IN   the interface name (eth0/lo)
// rec  OUT
static void get_if_info(const char *const itf, my_if_t * const rec) {
	struct ifreq ifr;
	int fd;

	strncpy(rec->ifname, itf, IFNAMSIZ);

	// We use this since root privileges are not needed and it does
	// not matter whether it is a UDP or a RAW socket to get data
	// on the interface
	fd = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);

	memset(&ifr, '\0', sizeof(ifr));
	strncpy(ifr.ifr_name, itf, IFNAMSIZ);
	if (-1 == ioctl(fd, SIOCGIFMTU, &ifr)) {
		Dprintf(L_FATAL, "cannot get the MTU of %s (SIOCGIFMTU)",
		    rec->ifname);
		exit(1);
	}
	rec->ifmtu = ifr.ifr_mtu;

	memset(&ifr, '\0', sizeof(ifr));
	strncpy(ifr.ifr_name, itf, IFNAMSIZ);
	if (-1 == ioctl(fd, SIOCGIFADDR, &ifr)) {
		Dprintf(L_FATAL, "cannot get the main IP address of %s (SIOCGIFADDR)", rec->ifname);
		exit(1);
	}
	rec->ifaddr =
	    ((struct sockaddr_in *) &(ifr.ifr_addr))->sin_addr.s_addr;

	memset(&ifr, '\0', sizeof(ifr));
	strncpy(ifr.ifr_name, itf, IFNAMSIZ);
	if (-1 == ioctl(fd, SIOCGIFINDEX, &ifr)) {
		Dprintf(L_FATAL, "cannot get the interface index of %s (SIOCGIFINDEX)", rec->ifname);
		exit(1);
	}
	rec->ifindex = ifr.ifr_ifindex;

	memset(&ifr, '\0', sizeof(ifr));
	strncpy(ifr.ifr_name, itf, IFNAMSIZ);
	if (-1 == ioctl (fd, SIOCGIFHWADDR, &ifr)) {
		Dprintf(L_FATAL, "cannot get the HW address of %s (SIOCGIFHWADDR)", rec->ifname);
		exit(1);
	}
	memcpy(rec->ifhwaddr, ifr.ifr_hwaddr.sa_data, sizeof(rec->ifhwaddr));

	
	close(fd);
	Dprintf(D_NFSP_UDPSPOOF, "interface %s : IP=%X MTU=%d index=%d", rec->ifname, rec->ifaddr,
	     rec->ifmtu, rec->ifindex);
}


void udpspoof_init() {
	get_if_info("lo", &if_lo);
	get_if_info("eth0", &if_eth0);
	eth0addr = if_eth0.ifaddr;
}



// Parameters:
// 
//  pseudoip: a struct iphdr followed by an struct udphdr UDP and the data
//            - UDP packet MUST be complete (esp. size and checksum)
//            - IP pseudo header MUST be of size sizeof(struct iphdr) and have
//            its source and address field set
//            
//  s: an IPPROTO_RAW socket
//
// WARNING:
//  This send is destructive on the buffer pointed to by pseudoip.
//  IOW, DO NOT USE THE BUFFER AFTER THIS FUNCTION.
//
void IP_send(struct iphdr * const pseudoip, const int s) {
	char *data;		// data to send
	struct udphdr *udp;	// original udp header
	struct iphdr *ip;	// moving ip header
	int datalen;		// length of data to send (type may be ushort)
	int frag;		// data from 0 to frag have been sent
	int tmp;		// scratch value
	int MTU;		// MTU used for transmission (8bytes multiple IIRC)
	int sent;		// nb of bytes effectively sent
	sockaddr_in_t destsock;
	socklen_t destsock_len = sizeof(sockaddr_in_t);

	udp = (struct udphdr *) ((char *) pseudoip + sizeof(struct iphdr));
	data = (char *) udp;
	frag = 0;
	datalen = ntohs(udp->len);

	pseudoip->version = 4;	// IPv4
	pseudoip->ihl = sizeof(struct iphdr) / 4;	// Number of 4bytes words
	pseudoip->tos = 0;	// Nothing special
	pseudoip->ttl = 32;	// A TTL
	pseudoip->id = htons(get_ipid());	// We set an IP id
	// pseudoip->{saddr,daddr,protocol==IPPROTO_UDP} are set
	// pseudoip->check will be filled by kernel

	memset(&destsock, '\0', sizeof(sockaddr_t));
	destsock.sin_family = AF_INET;
	destsock.sin_port = ntohs(udp->source);

	// If pseudoip->daddr is the address of an interface on _THIS_
	// machine then we set this address to the localhost interface
	// else we set it the normal way
	if (pseudoip->daddr == if_eth0.ifaddr) {
		// most likely 127.0.0.1
		destsock.sin_addr.s_addr = if_lo.ifaddr;
		MTU = if_lo.ifmtu;
	} else {
		// else the following line should be used
		destsock.sin_addr.s_addr = ntohl(pseudoip->daddr);
		MTU = if_eth0.ifmtu;
	}
	// The interface mtu will be modified so that
	//    ( MTU - sizeof(struct ddiphdr) ) % 8 == 0
	MTU = sizeof(struct iphdr) + ((MTU - sizeof(struct iphdr)) & 0xFFF8);

	// We don't use inet_ntoa since it is not MT-safe
	do {
#define tmps (pseudoip->saddr)
#define tmpd (pseudoip->daddr)
		Dprintf(D_NFSP_UDPSPOOF, "UDP: about to send packet from %d.%d.%d.%d:%d to %d.%d.%d.%d:%d MTU=%d",
		       	(tmps >> 0) & 0xFF, (tmps >> 8) & 0xFF, (tmps >> 16) & 0xFF, (tmps >> 24) & 0xFF,
		   	ntohs(udp->source),
		   	(tmpd >> 0) & 0xFF, (tmpd >> 8) & 0xFF, (tmpd >> 16) & 0xFF, (tmpd >> 24) & 0xFF,
		   	ntohs(udp->dest), MTU);
#undef tmps
#undef tmpd
	} while (0);

	do {
		ip = (struct iphdr *) (data + frag - sizeof(struct iphdr));

		(void) memmove(ip, pseudoip, sizeof(struct iphdr));

		if (datalen - frag > MTU - sizeof(struct iphdr)) {
			// Fraggy happy people...
			tmp = MTU;
			ip->tot_len = htons(tmp);
			ip->frag_off =
			    htons(((frag / 8) & IP_OFFMASK) | IP_MF);

		} else {
			// Last packet
			tmp = datalen - frag + sizeof(struct iphdr);
			ip->tot_len = htons(tmp);
			ip->frag_off = htons(((frag / 8) & IP_OFFMASK));
		}

		sent = sendto(s, (char *) ip, tmp, 0,
			   (sockaddr_t *) & destsock, destsock_len);
		if (-1 == sent) {
			Dprintf(L_WARNING, "cannot sendto() errno=%d", errno);
		}

		Dprintf(D_NFSP_UDPSPOOF, "%dB / %dB were sent (IPhdr: %dB  UDPhdr: %dB)",
				sent, tmp, sizeof(struct iphdr), sizeof(struct udphdr));

		frag += tmp - sizeof(struct iphdr);

	} while (frag < datalen);
}


// Parameters:
// 
//  pseudoip: a struct iphdr followed by an struct udphdr UDP and the data
//            - UDP packet MUST be complete (esp. size and checksum)
//            - IP pseudo header MUST be of size sizeof(struct iphdr) and have
//            its source and address field set
//            
//  dest_mac_addr: MAC address of the destination host (matching with the IP
//  address set in destination field)
//
//  s: an open SOCK_DGRAM socket with ETH_P_ALL (link layer level socket - man 7 packet)
//
// WARNING:
//  This send is destructive on the buffer pointed to by pseudoip.
//  IOW, DO NOT USE THE BUFFER AFTER THIS FUNCTION.
//
void ETHERNET_send(struct iphdr * const pseudoip, const int s, const char * const dest_mac_addr) {
	char *data;		// data to send
	struct udphdr *udp;	// original udp header
	struct iphdr *ip;	// moving ip header
	int datalen;		// length of data to send (type may be ushort)
	int frag;		// data from 0 to frag have been sent
	int tmp;		// scratch value
	int MTU;		// MTU used for transmission (8bytes multiple IIRC)
	int sent;		// nb of bytes effectively sent
	struct sockaddr_ll destsock;
	socklen_t destsock_len = sizeof(destsock);

	udp = (struct udphdr *) ((char *) pseudoip + sizeof(struct iphdr));
	data = (char *) udp;
	frag = 0;
	datalen = ntohs(udp->len);

	// TODO why not assign these fields in the do_ip_header function ?
	pseudoip->version = 4;				// IPv4
	pseudoip->ihl = sizeof(struct iphdr) / 4;	// Number of 4bytes words
	pseudoip->tos = 0;				// Nothing special
	pseudoip->ttl = 32;				// A TTL
	pseudoip->id = htons(get_ipid());		// We set an IP id
	
	// pseudoip->{saddr,daddr,protocol==IPPROTO_UDP} are already be set by our caller

	// Note: the checksum is set to 0 but has to be recomputed for each IP packet
	pseudoip->check = 0;

	memset(&destsock, '\0', sizeof(destsock));
	destsock.sll_family = AF_PACKET;
	destsock.sll_halen = 6;			// HW address length (MAC)
	destsock.sll_protocol = htons(ETH_P_IP);


	Dprintf(D_NFSP_UDPSPOOF, "if_eth0.ifindex=%d\n", if_eth0.ifindex);

	// TODO optimize with the real ifindex (see man 7 packet).

	// We do some minimal routing here:
	//  - if pseudoip->daddr belongs to this machine then we route through 'lo'
	//  - else we route through 'eth0'
	if (pseudoip->daddr != if_eth0.ifaddr) {
		destsock.sll_ifindex = if_eth0.ifindex;
		memcpy(destsock.sll_addr, dest_mac_addr, 6);
		MTU = if_eth0.ifmtu;
	} else {
#warning "ETHERNET_send(): sending through 'lo' does not work"
		Dprintf(L_WARNING, "trying to send through 'lo' - known bug");
		destsock.sll_ifindex = if_lo.ifindex;
		memcpy(destsock.sll_addr, if_lo.ifhwaddr, 6);
		MTU = if_lo.ifmtu;
	}

	Dprintf(D_NFSP_UDPSPOOF, "sending through interface %d", destsock.sll_ifindex);

	// The interface mtu will be modified so that
	//    ( MTU - sizeof(struct ddiphdr) ) % 8 == 0
	MTU = sizeof(struct iphdr) + ((MTU - sizeof(struct iphdr)) & 0xFFF8);

	// We don't use inet_ntoa since it is not MT-safe
	do {
#define tmps (pseudoip->saddr)
#define tmpd (pseudoip->daddr)
#define macd (destsock.sll_addr)
		Dprintf(D_NFSP_UDPSPOOF,
			"UDP: about to send packet from %d.%d.%d.%d:%d to %d.%d.%d.%d:%d "
			"(%2.2x:%2.2x:%2.2x:%2.2x:%2.2x:%2.2x) MTU=%d",
		       	(tmps >> 0) & 0xFF, (tmps >> 8) & 0xFF, (tmps >> 16) & 0xFF, (tmps >> 24) & 0xFF,
		   	ntohs(udp->source),
		   	(tmpd >> 0) & 0xFF, (tmpd >> 8) & 0xFF, (tmpd >> 16) & 0xFF, (tmpd >> 24) & 0xFF,
		   	ntohs(udp->dest),
			macd[0] & 0xFF, macd[1] & 0xFF, macd[2] & 0xFF,
		       	macd[3] & 0xFF, macd[4] & 0xFF, macd[5] & 0xFF,
			MTU);
#undef macd
#undef tmps
#undef tmpd
	} while (0);

	do {
		ip = (struct iphdr *) (data + frag - sizeof(struct iphdr));

		(void) memmove(ip, pseudoip, sizeof(struct iphdr));

		// Handle the IP fragmentation
		if (datalen - frag > MTU - sizeof(struct iphdr)) {
			// Fraggy happy people...
			tmp = MTU;
			ip->tot_len = htons(tmp);
			ip->frag_off =
			    htons(((frag / 8) & IP_OFFMASK) | IP_MF);

		} else {
			// Last packet
			tmp = datalen - frag + sizeof(struct iphdr);
			ip->tot_len = htons(tmp);
			ip->frag_off = htons(((frag / 8) & IP_OFFMASK));
		}

		// The packet is ready (all headers are set) so we compute the
		// IP checksum of the packet (header checksum)
		ip->check = 0;
		ip->check = in_cksum((unsigned short *) ip, sizeof(struct iphdr));
	
		sent = sendto(s, (char *) ip, tmp, 0,
			   (sockaddr_t *) & destsock, destsock_len);
		if (-1 == sent) {
			Dprintf(L_WARNING, "cannot sendto() errno=%d", errno);
		}

		Dprintf(D_NFSP_UDPSPOOF, "%dB / %dB were sent (IPhdr: %dB  UDPhdr: %dB)",
				sent, tmp, sizeof(struct iphdr), sizeof(struct udphdr));

		frag += tmp - sizeof(struct iphdr);

	} while (frag < datalen);
}


void UDP_send(struct iphdr * const pseudoip, const int s) {
	struct udphdr *udp;	// original udp header
	char *payload;
	struct sockaddr_in sin;
	int sent;

	udp = (struct udphdr *) ((char *) pseudoip + sizeof(struct iphdr));
	payload = (char *) udp + sizeof (struct udphdr);

	memset(&sin, '\0', sizeof(sin));
	sin.sin_family = AF_INET;
	sin.sin_addr.s_addr = pseudoip->daddr;
	sin.sin_port = udp->dest;

	// We don't use inet_ntoa since it is not MT-safe
	do {
#define tmpd (pseudoip->daddr)
		Dprintf(D_NFSP_UDPSPOOF, "UDP: about to send packet to %d.%d.%d.%d:%d (plain UDP sendto)",
		   	(tmpd >> 0) & 0xFF, (tmpd >> 8) & 0xFF, (tmpd >> 16) & 0xFF, (tmpd >> 24) & 0xFF,
		   	ntohs(udp->dest));
#undef tmps
	} while (0);


	sent = sendto(s, payload, ntohs(udp->len) - sizeof(struct udphdr), 0, &sin, sizeof(sin));
	if (sent == -1) {
		Dprintf(L_WARNING, "sendto failed errno=%d", errno);
		return;
	}
	Dprintf(D_NFSP_UDPSPOOF, "%dB data bytes were sent (add IPhdr: %dB  UDPhdr: %dB for on wire size)",
				sent, sizeof(struct iphdr), sizeof(struct udphdr));
}


